"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from .models import BltnModels
from .period import DriverBltnPeriod
from .regexp import DriverBltnRegexp
from .store import DriverBltnStore



__all__ = [
    'BltnModels',
    'DriverBltnStore',
    'DriverBltnPeriod',
    'DriverBltnRegexp']



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

from pytest import raises

from ..helpers import request_action
from ..origin import HubiOrigin
from ...utils import Idempotent

if TYPE_CHECKING:
    from ...homie import Homie
    from ...utils import TestBodies



def test_request_action(
    homie: 'Homie',
    bodies: 'TestBodies',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    :param bodies: Locations and groups for use in testing.
    """

    childs = homie.childs
    origins = childs.origins
    devices = childs.devices
    scenes = childs.scenes

    scene1 = scenes['active']
    scene2 = scenes['standby']


    planets = bodies.planets

    for planet in planets:

        origin = origins[
            f'{planet}_hubitat']

        assert isinstance(
            origin, HubiOrigin)

        light = devices[
            f'{planet}_light2']

        assert origin.refresh()


        request_action(
            origin=origin,
            target=light,
            scene=scene1)

        with raises(Idempotent):

            request_action(
                origin=origin,
                target=light,
                scene=scene2)

        request_action(
            origin=origin,
            target=light,
            scene=scene1,
            change=False)


        request_action(
            origin=origin,
            target=light,
            state='nopower')

        with raises(Idempotent):

            request_action(
                origin=origin,
                target=light,
                state='poweron')

        request_action(
            origin=origin,
            target=light,
            state='nopower',
            change=False)


        request_action(
            origin=origin,
            target=light,
            color='ff00cc')

        with raises(Idempotent):

            request_action(
                origin=origin,
                target=light,
                color='003333')

        request_action(
            origin=origin,
            target=light,
            color='ff00cc',
            change=False)


        request_action(
            origin=origin,
            target=light,
            level=100)

        with raises(Idempotent):

            request_action(
                origin=origin,
                target=light,
                level=20)

        request_action(
            origin=origin,
            target=light,
            level=100,
            change=False)



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ...homie import Homie



def test_HomiePlugin_cover(
    homie: 'Homie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    """

    # Remove if not needed



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from copy import deepcopy
from dataclasses import dataclass
from typing import TYPE_CHECKING

from encommon.types import DictStrAny

from ..homie.threads import HomieUpdate
from ..homie.threads import HomieUpdateItem

if TYPE_CHECKING:
    from ..homie.childs import HomieOrigin



@dataclass
class HubiUpdateItem(HomieUpdateItem):
    """
    Contain information for sharing using the Python queue.
    """

    fetch: list[DictStrAny]


    def __init__(
        self,
        origin: 'HomieOrigin',
        fetch: list[DictStrAny],
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        fetch = deepcopy(fetch)

        self.fetch = fetch

        super().__init__(origin)



class HubiUpdate(HomieUpdate):
    """
    Common methods and routines for Homie Automate threads.
    """



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING
from typing import Type

if TYPE_CHECKING:
    from .common import HomiePluginKinds
    from .driver import HomieDriver
    from ..common import HomieFamily
    from ..homie import Homie
    from ..params import HomiePluginParams



class HomiePlugin:
    """
    Match specific conditions for determining desired state.

    :param homie: Primary class instance for Homie Automate.
    :param params: Parameter options for plugins operations.
    """

    __homie: 'Homie'

    __params: 'HomiePluginParams'
    __drivers: list['HomieDriver']


    def __init__(
        self,
        homie: 'Homie',
        params: 'HomiePluginParams',
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        self.__homie = homie
        self.__params = params

        self.__drivers = []

        self.__build_drivers()


    def __build_drivers(
        self,
    ) -> None:
        """
        Construct instances using the configuration parameters.
        """

        params = self.params

        objects = self.get_drivers()


        drivers: list['HomieDriver'] = []


        items = (
            params.__dict__
            .items())

        for name, driver in items:

            if name not in objects:
                continue

            if driver is None:
                continue

            _object = objects[name]

            object = (
                _object(self, driver))

            drivers.append(object)


        self.__drivers = drivers


    def validate(
        self,
    ) -> None:
        """
        Perform advanced validation on the parameters provided.
        """

        drivers = self.__drivers

        for driver in drivers:
            driver.validate()


    @property
    def homie(
        self,
    ) -> 'Homie':
        """
        Return the Homie instance to which this instance belongs.

        :returns: Homie instance to which this instance belongs.
        """

        return self.__homie


    @property
    def family(
        self,
    ) -> 'HomieFamily':
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return 'builtins'


    @property
    def kind(
        self,
    ) -> 'HomiePluginKinds':
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        raise NotImplementedError


    @property
    def params(
        self,
    ) -> 'HomiePluginParams':
        """
        Return the Pydantic model containing the configuration.

        :returns: Pydantic model containing the configuration.
        """

        return self.__params


    @property
    def drivers(
        self,
    ) -> list['HomieDriver']:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return list(self.__drivers)


    def get_drivers(
        self,
    ) -> dict[str, Type['HomieDriver']]:
        """
        Return the Homie class definition for its instantiation.

        :returns: Homie class definition for its instantiation.
        """

        raise NotImplementedError



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from ..models import HomieModels



def test_HomieModels_cover() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    models = HomieModels

    assert models.homie()
    assert models.printer()
    assert models.service()

    assert models.builtins()
    assert models.hubitat()
    assert models.philips()
    assert models.ubiquiti()

    assert models.child()
    assert models.origin()
    assert models.device()
    assert models.group()
    assert models.scene()
    assert models.desire()
    assert models.aspire()
    assert models.plugin()
    assert models.occur()
    assert models.where()

    assert models.stage()
    assert models.queue()
    assert models.thread()
    assert models.action()
    assert models.update()
    assert models.stream()
    assert models.desired()
    assert models.aspired()



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Optional
from typing import TYPE_CHECKING
from typing import Type

from encommon.types import NCNone

from .member import HomieMember
from ...hubitat import HubiUpdate
from ...philips import PhueUpdate
from ...ubiquiti import UbiqUpdate

if TYPE_CHECKING:
    from ..childs import HomieOrigin
    from ..threads import HomieThread



class HomieUpdates(HomieMember):
    """
    Common methods and routines for Homie Automate members.
    """


    def operate(
        self,
    ) -> None:
        """
        Perform the operation related to Homie service members.
        """

        self.operate_streams()


    def operate_streams(
        self,
    ) -> None:
        """
        Perform the operation related to Homie service members.
        """

        vacate = self.vacate
        squeue = self.squeue

        threads = (
            self.threads
            .values())


        def _put_update() -> None:

            name = sitem.origin
            origin = thread.origin
            _name = origin.name

            if name != _name:
                return None

            (thread
             .squeue.put(sitem))


        while not squeue.empty:

            sitem = squeue.get()

            if vacate.is_set():
                continue

            for thread in threads:
                _put_update()


    def get_thread(  # noqa: CFQ004
        self,
        origin: 'HomieOrigin',
    ) -> Optional[Type['HomieThread']]:
        """
        Return the Homie class definition for its instantiation.

        :param origin: Child class instance for Homie Automate.
        :returns: Homie class definition for its instantiation.
        """

        family = origin.family

        if family == 'hubitat':
            return HubiUpdate

        if family == 'philips':
            return PhueUpdate

        if family == 'ubiquiti':
            return UbiqUpdate

        return NCNone



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from argparse import ArgumentParser
from sys import argv as sys_argv
from typing import Optional

from encommon.times import Times
from encommon.types import DictStrAny

from ..homie import Homie
from ..homie import HomieConfig



def arguments(
    args: Optional[list[str]] = None,
) -> DictStrAny:
    """
    Construct arguments which are associated with the file.

    :param args: Override the source for the main arguments.
    :returns: Construct arguments from command line options.
    """

    parser = ArgumentParser()

    args = args or sys_argv[1:]


    parser.add_argument(
        '--config',
        required=True,
        help=(
            'complete or relative '
            'path to config file'))


    parser.add_argument(
        '--console',
        action='store_true',
        default=False,
        help=(
            'write log messages '
            'to standard output'))


    parser.add_argument(
        '--debug',
        action='store_true',
        default=False,
        help=(
            'increase logging level '
            'for standard output'))


    parser.add_argument(
        '--dry-run',
        action='store_true',
        # Argument is always true
        default=True,
        dest='dryrun',
        help='always true for script')


    parser.add_argument(
        '--print_desire',
        action='store_true',
        # Argument is always true
        default=True,
        dest='pdesire',
        help=(
            'print the desired '
            'state to console'))


    return vars(
        parser
        .parse_args(args))



def operation(
    # NOCVR
    homie: Homie,
) -> None:
    """
    Perform whatever operation is associated with the file.

    :param homie: Primary class instance for Homie Automate.
    """

    assert homie.refresh()

    time = Times('now')

    aitems = (
        homie.desired
        .items(time))

    for aitem in aitems:
        homie.printer(aitem)



def execution(
    # NOCVR
    args: Optional[list[str]] = None,
) -> None:
    """
    Perform whatever operation is associated with the file.

    :param args: Override the source for the main arguments.
    """

    config = HomieConfig(
        arguments(args))

    config.logger.start()

    homie = Homie(config)

    config.logger.log_i(
        base='execution/desired',
        status='started')

    operation(homie)

    config.logger.log_i(
        base='execution/desired',
        status='stopped')

    config.logger.stop()



if __name__ == '__main__':
    execution()  # NOCVR



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Annotated
from typing import Optional

from pydantic import Field

from .child import HomieChildParams
from .stage import HomieStageParams



class HomieSceneParams(HomieChildParams, extra='forbid'):
    """
    Process and validate the Homie configuration parameters.
    """

    label: Annotated[
        Optional[str],
        Field(None,
              description='Scene name in the origin',
              min_length=1)]

    stage: Annotated[
        HomieStageParams,
        Field(default_factory=HomieStageParams,
              description='Default device scene config')]

    devices: Annotated[
        Optional[dict[str, HomieStageParams]],
        Field(None,
              description='Device specific scene config',
              min_length=1)]



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Optional
from typing import TYPE_CHECKING

from encommon.times import Times
from encommon.types import getate
from encommon.types import sort_dict

if TYPE_CHECKING:
    from ..helpers import PhueFetch



_CHANGED = dict[str, Optional[Times]]

_SENSORS = dict[str, str]

_REPORTS = ['button', 'contact', 'motion']



def phue_changed(
    source: 'PhueFetch',
) -> _CHANGED:
    """
    Return the timestamp for the services that have changed.

    :param source: Content which will be shown after header.
    :returns: Timestamp for the services that have changed.
    """

    changed: _CHANGED = {}


    services = (
        source['services'])


    for item in services:

        rtype = item['rtype']

        if rtype not in _REPORTS:
            continue

        fetch = item['_source']


        base = f'{rtype}_report'
        key = 'changed'

        if rtype == 'button':
            key = 'updated'


        time = getate(
            (fetch[rtype]
             if rtype != 'contact'
             else fetch),
            f'{base}/{key}')


        index = getate(
            fetch,
            'metadata/control_id')

        if index is not None:
            rtype += str(index)

        changed[rtype] = (
            Times(time)
            if time is not None
            else None)


    return sort_dict(changed)



def phue_sensors(
    source: 'PhueFetch',
) -> _SENSORS:
    """
    Return the unique identifier for services on the device.

    :param source: Content which will be shown after header.
    :returns: Timestamp for the services that have changed.
    """

    sensors: _SENSORS = {}


    services = (
        source['services'])


    for item in services:

        rtype = item['rtype']
        rid = item['rid']

        if rtype not in _REPORTS:
            continue

        fetch = item['_source']

        index = getate(
            fetch,
            'metadata/control_id')

        if index is not None:
            rtype += str(index)

        sensors[rtype] = rid


    return sort_dict(sensors)



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Literal
from typing import TYPE_CHECKING
from typing import Type

from encommon.times import Times

from .plugin import HomiePlugin
from ..models import HomieModels
from ...utils import InvalidParam
from ...utils import UnexpectedCondition

if TYPE_CHECKING:
    from .driver import HomieDriver
    from ..params import HomieWhereParams



class HomieWhere(HomiePlugin):
    """
    Match specific conditions for determining desired state.
    """


    def validate(
        self,
    ) -> None:
        """
        Perform advanced validation on the parameters provided.
        """

        params = self.params

        _bltn_store = params.builtins_store
        _bltn_period = params.builtins_period
        _phue_change = params.philips_change
        _ubiq_client = params.ubiquiti_client

        super().validate()


        enabled = [
            1 if _bltn_store else 0,
            1 if _bltn_period else 0,
            1 if _phue_change else 0,
            1 if _ubiq_client else 0]

        if sum(enabled) != 1:

            raise InvalidParam(
                error='minimal')


    @property
    def kind(
        self,
    ) -> Literal['where']:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return 'where'


    @property
    def params(
        self,
    ) -> 'HomieWhereParams':
        """
        Return the Pydantic model containing the configuration.

        :returns: Pydantic model containing the configuration.
        """

        model = (
            HomieModels
            .where())

        params = super().params

        assert isinstance(
            params, model)

        return params


    def get_drivers(
        self,
    ) -> dict[str, Type['HomieDriver']]:
        """
        Return the Homie class definition for its instantiation.

        :returns: Homie class definition for its instantiation.
        """

        from ...builtins import (
            DriverBltnPeriod)

        from ...builtins import (
            DriverBltnStore)

        from ...philips import (
            DriverPhueChange)

        from ...ubiquiti import (
            DriverUbiqClient)


        return {

            'builtins_period': (
                DriverBltnPeriod),

            'builtins_store': (
                DriverBltnStore),

            'philips_change': (
                DriverPhueChange),

            'ubiquiti_client': (
                DriverUbiqClient)}


    def when(
        self,
        time: Times,
    ) -> bool:
        """
        Return the boolean indicating the conditional outcomes.

        .. note::
           Somewhat similar to same method within HomieOccur.

        :param time: Time that will be used in the conditionals.
        :returns: Boolean indicating the conditional outcomes.
        """

        drivers = self.drivers

        wheres: list[bool] = []


        for driver in drivers:

            where = driver.where(time)

            wheres.append(where)


        if len(wheres) >= 1:
            return all(wheres)

        raise UnexpectedCondition



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from ..models import PhueModels



def test_PhueModels_cover() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    models = PhueModels

    assert models.origin()
    assert models.action()
    assert models.update()
    assert models.stream()

    drivers = models.drivers()

    assert drivers.button()
    assert drivers.change()
    assert drivers.contact()
    assert drivers.motion()



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from .aspire import HomieAspire
from .child import HomieChild
from .common import HomieChildKinds
from .desire import HomieDesire
from .device import HomieDevice
from .group import HomieGroup
from .homie import HomieChilds
from .origin import HomieOrigin
from .scene import HomieScene



__all__ = [
    'HomieChild',
    'HomieChildKinds',
    'HomieChilds',
    'HomieOrigin',
    'HomieDevice',
    'HomieGroup',
    'HomieScene',
    'HomieAspire',
    'HomieDesire']



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Annotated
from typing import Optional

from encommon.config import Params
from encommon.types import BaseModel

from pydantic import Field

from .aspire import HomieAspireParams
from .desire import HomieDesireParams
from .device import HomieDeviceParams
from .group import HomieGroupParams
from .origin import HomieOriginParams
from .scene import HomieSceneParams
from .service import HomieServiceParams



class HomiePrinterParams(BaseModel, extra='forbid'):
    """
    Process and validate the Homie configuration parameters.
    """

    action: Annotated[
        bool,
        Field(False,
              description='Print the actions to console')]

    update: Annotated[
        bool,
        Field(False,
              description='Print the updates to console')]

    stream: Annotated[
        bool,
        Field(False,
              description='Print the streams to console')]

    desire: Annotated[
        bool,
        Field(False,
              description='Print the aspires to console')]

    aspire: Annotated[
        bool,
        Field(False,
              description='Print the aspires to console')]



class HomieParams(Params, BaseModel, extra='forbid'):
    """
    Process and validate the core configuration parameters.
    """

    database: Annotated[
        str,
        Field('sqlite:///:memory:',
              description='Database connection string',
              min_length=1)]

    dryrun: Annotated[
        bool,
        Field(False,
              description='Determine if changes are made')]

    potent: Annotated[
        bool,
        Field(True,
              description='Ignore idempotency in change')]

    printer: Annotated[
        HomiePrinterParams,
        Field(default_factory=HomiePrinterParams,
              description='Print the stream to console')]

    service: Annotated[
        HomieServiceParams,
        Field(default_factory=HomieServiceParams,
              description='Parameters for Homie Service')]

    origins: Annotated[
        Optional[dict[str, HomieOriginParams]],
        Field(None,
              description='Parameters for Homie Automate',
              min_length=1)]

    devices: Annotated[
        Optional[dict[str, HomieDeviceParams]],
        Field(None,
              description='Parameters for Homie Automate',
              min_length=1)]

    groups: Annotated[
        Optional[dict[str, HomieGroupParams]],
        Field(None,
              description='Parameters for Homie Automate',
              min_length=1)]

    scenes: Annotated[
        Optional[dict[str, HomieSceneParams]],
        Field(None,
              description='Parameters for Homie Automate',
              min_length=1)]

    desires: Annotated[
        Optional[dict[str, HomieDesireParams]],
        Field(None,
              description='Parameters for Homie Automate',
              min_length=1)]

    aspires: Annotated[
        Optional[dict[str, HomieAspireParams]],
        Field(None,
              description='Parameters for Homie Automate',
              min_length=1)]



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from .action import PhueAction
from .device import PhueDevice
from .models import PhueModels
from .origin import PhueOrigin
from .plugins import DriverPhueButton
from .plugins import DriverPhueChange
from .plugins import DriverPhueContact
from .plugins import DriverPhueMotion
from .plugins import DriverPhueScene
from .stream import PhueStream
from .update import PhueUpdate



__all__ = [
    'PhueOrigin',
    'PhueDevice',
    'PhueAction',
    'PhueUpdate',
    'PhueStream',
    'PhueModels',
    'DriverPhueButton',
    'DriverPhueChange',
    'DriverPhueContact',
    'DriverPhueMotion',
    'DriverPhueScene']



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

from ..action import PhueAction
from ..action import PhueActionItem

if TYPE_CHECKING:
    from ...homie import HomieService
    from ...utils import TestBodies



def test_PhueAction_cover(
    service: 'HomieService',
    bodies: 'TestBodies',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param service: Ancilary Homie Automate class instance.
    :param bodies: Locations and groups for use in testing.
    """

    homie = service.homie
    childs = homie.childs
    origins = childs.origins
    devices = childs.devices
    groups = childs.groups

    actions = service.actions
    threads = actions.threads

    model = PhueActionItem


    planets = bodies.planets
    moons = dict(bodies.moons)

    for planet in planets:

        moon = moons[planet]

        origin = origins[
            f'{planet}_philips']

        device = devices[
            f'{planet}_light1']

        group = groups[moon]

        assert origin.refresh()

        thread = threads[
            f'{planet}_philips']

        assert isinstance(
            thread, PhueAction)


        aitem = model(
            origin, device)

        thread.execute(aitem)


        aitem.device = None

        thread.execute(aitem)


        aitem = model(
            origin, group)

        group.params.origin = None

        thread.execute(aitem)



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from dataclasses import dataclass
from typing import Optional
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..childs import HomieAspire
    from ..childs import HomieScene
    from ..common import HomieState
    from ..homie import Homie
    from ..threads import HomieActionNode
    from ..threads import HomieStreamItem



@dataclass
class HomieAspiredItem:
    """
    Contain information related to the aspired child state.
    """

    target: 'HomieActionNode'

    aspire: 'HomieAspire'

    state: Optional['HomieState'] = None
    color: Optional[str] = None
    level: Optional[int] = None
    scene: Optional['HomieScene'] = None


    def __init__(
        self,
        target: 'HomieActionNode',
        aspire: 'HomieAspire',
    ) -> None:
        """
        Initialize instance for class using provided parameters.

        .. note::
           Somewhat similar to same method in HomieDesiredItem.
        """

        homie = aspire.homie
        childs = homie.childs

        params = aspire.params

        self.target = target

        self.aspire = aspire

        scene = params.scene
        stage = params.stage

        self.state = None
        self.color = None
        self.level = None
        self.scene = None


        if scene is not None:

            _scene = (
                childs.scenes
                [scene])

            self.scene = _scene


        if stage is not None:

            _state = stage.state
            _color = stage.color
            _level = stage.level

            self.state = _state
            self.color = _color
            self.level = _level


    def __lt__(
        self,
        other: 'HomieAspiredItem',
    ) -> bool:
        """
        Built-in method for comparing this instance with another.

        .. note::
           Useful with sorting to influence consistent output.

        :param other: Other value being compared with instance.
        :returns: Boolean indicating outcome from the operation.
        """

        name = self.target.name
        _name = other.target.name

        return bool(name < _name)



_ASPIRED = list[HomieAspiredItem]
_MATCHED = set['HomieAspire']



class HomieAspired:
    """
    Contain information related to the aspired child state.

    :param homie: Primary class instance for Homie Automate.
    """

    __homie: 'Homie'


    def __init__(
        self,
        homie: 'Homie',
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        self.__homie = homie


    def items(
        self,
        sitem: 'HomieStreamItem',
    ) -> _ASPIRED:
        """
        Return the action state objects for the related targets.

        ..note::
          Somewhat similar to same method within HomieDesire.

        :param sitem: Item containing information for operation.
        :returns: Action state objects for the related targets.
        """

        homie = self.__homie
        persist = homie.persist
        childs = homie.childs

        aspires = (
            childs.aspires
            .values())

        aspired: _ASPIRED = []
        matched: _MATCHED = set()

        model = HomieAspiredItem


        for aspire in aspires:

            when = aspire.when(sitem)

            if when is False:
                continue

            matched.add(aspire)

            if aspire.paused:

                homie.logger.log_d(
                    base=self,
                    item=aspire,
                    name=aspire,
                    status='pause')

                continue

            homie.logger.log_d(
                base=self,
                item=aspire,
                name=aspire,
                status='match')

            aspire.matched()


        for aspire in matched:

            store = (
                aspire.params
                .store)

            if store is None:
                continue

            if homie.dryrun:
                continue

            for value in store:

                persist.insert(
                    value.unique,
                    value.value,
                    value.expire)


        for aspire in matched:

            devices = aspire.devices

            for device in devices:

                object = model(
                    device, aspire)

                aspired.append(object)

            groups = aspire.groups

            for group in groups:

                object = model(
                    group, aspire)

                aspired.append(object)


        return list(aspired)



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from .config import HomieConfig
from .homie import Homie
from .service import HomieService



__all__ = [
    'Homie',
    'HomieConfig',
    'HomieService']



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

from encommon.types import inrepr
from encommon.types import instr
from encommon.types import lattrs

if TYPE_CHECKING:
    from ...homie import Homie
    from ....utils import TestBodies



def test_HomieScene(
    homie: 'Homie',
    bodies: 'TestBodies',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    :param bodies: Locations and groups for use in testing.
    """

    childs = homie.childs
    origins = childs.origins
    devices = childs.devices
    scenes = childs.scenes


    scene = scenes['standby']


    attrs = lattrs(scene)

    assert attrs == [
        '_HomieChild__homie',
        '_HomieChild__name',
        '_HomieChild__params']


    assert inrepr(
        'scene.HomieScene',
        scene)

    assert hash(scene) > 0

    assert instr(
        'scene.HomieScene',
        scene)


    scene.validate()

    assert scene.homie

    assert scene.name == 'standby'

    assert scene.family == 'builtins'

    assert scene.kind == 'scene'

    assert scene.params

    assert scene.dumped

    assert scene.stage()


    planets = bodies.planets

    for planet in planets:

        origin = origins[
            f'{planet}_philips']

        origin.refresh()

        assert scene.source(origin)


        device = devices[
            f'{planet}_light1']

        stage1, stage2 = (
            scene.stage(device),
            scene.stage())

        assert stage1 == stage2


        device = devices[
            f'{planet}_light2']

        stage1, stage2 = (
            scene.stage(device),
            scene.stage())

        assert stage1 != stage2



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Annotated

from encommon.types import BaseModel

from enconnect.ubiquiti import RouterParams

from pydantic import Field



class UbiqOriginParams(BaseModel, extra='forbid'):
    """
    Process and validate the Homie configuration parameters.
    """

    router: Annotated[
        RouterParams,
        Field(...,
              description='Connection specific parameters')]



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from json import loads
from typing import TYPE_CHECKING

from encommon.types import DictStrAny
from encommon.types import lattrs
from encommon.utils import load_sample
from encommon.utils import prep_sample
from encommon.utils import read_text
from encommon.utils import rvrt_sample
from encommon.utils.sample import ENPYRWS

from . import SAMPLES
from ..origin import PhueOrigin
from ..update import PhueUpdateItem

if TYPE_CHECKING:
    from ...homie import Homie
    from ...homie import HomieService
    from ...utils import TestBodies



def test_PhueOrigin(
    homie: 'Homie',
    bodies: 'TestBodies',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    :param bodies: Locations and groups for use in testing.
    """

    childs = homie.childs
    origins = childs.origins


    planets = bodies.planets

    family = 'philips'

    for planet in planets:

        name = f'{planet}_{family}'

        origin = origins[name]

        assert isinstance(
            origin, PhueOrigin)


        attrs = lattrs(origin)

        assert attrs == [
            '_HomieChild__homie',
            '_HomieChild__name',
            '_HomieChild__params',
            '_PhueOrigin__bridge',
            '_PhueOrigin__fetch',
            '_PhueOrigin__merge']


        origin.validate()

        assert origin.homie

        assert origin.name == name

        assert origin.family == family

        assert origin.kind == 'origin'

        assert origin.params

        assert origin.dumped


        assert origin.bridge

        assert not origin.fetch

        assert not origin.merge


        assert origin.refresh()

        assert origin.fetch

        assert origin.merge



def test_PhueOrigin_action(
    homie: 'Homie',
    service: 'HomieService',
    bodies: 'TestBodies',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    :param service: Ancilary Homie Automate class instance.
    :param bodies: Locations and groups for use in testing.
    """

    childs = homie.childs
    origins = childs.origins
    devices = childs.devices
    groups = childs.groups
    scenes = childs.scenes

    scene1 = scenes['active']
    scene2 = scenes['standby']


    planets = bodies.planets
    moons = dict(bodies.moons)

    for planet in planets:

        origin = origins[
            f'{planet}_philips']

        assert isinstance(
            origin, PhueOrigin)

        assert origin.refresh()


        moon = moons[planet]

        device = devices[
            f'{planet}_light1']

        group = groups[moon]


        aitem = (
            origin.get_action(
                target=device,
                scene=scene2))

        assert not (
            origin.set_action(
                device, aitem))


        aitem = (
            origin.get_action(
                target=group,
                scene=scene2))

        assert not (
            origin.set_action(
                group, aitem))


        aitem = (
            origin.get_action(
                target=device,
                state='nopower',
                color='ff00cc',
                level=100))

        assert (
            origin.set_action(
                device, aitem))

        assert (
            origin.set_action(
                device, aitem,
                change=False))


        aitem = (
            origin.get_action(
                target=device,
                state='poweron',
                color='003333',
                level=80))

        assert not (
            origin.set_action(
                device, aitem))


        aitem = (
            origin.get_action(
                target=group,
                scene=scene1.name))

        assert (
            origin.set_action(
                group, aitem,
                change=False))



def test_PhueOrigin_update(
    homie: 'Homie',
    service: 'HomieService',
    replaces: DictStrAny,
    bodies: 'TestBodies',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    :param service: Ancilary Homie Automate class instance.
    :param replaces: Mapping of what to replace in samples.
    :param bodies: Locations and groups for use in testing.
    """

    childs = homie.childs
    origins = childs.origins
    member = service.updates
    uqueue = member.uqueue

    samples = SAMPLES / 'stream'


    planets = bodies.planets

    for planet in planets:

        origin = origins[
            f'{planet}_philips']

        assert isinstance(
            origin, PhueOrigin)

        assert origin.refresh()

        fetch = origin.fetch


        uitem = origin.get_update()

        assert isinstance(
            uitem, PhueUpdateItem)

        uitem.fetch['foo'] = 'bar'

        origin.set_update(uitem)

        origin.put_update(uqueue)

        _fetch = origin.fetch


        assert _fetch != fetch


        source = read_text(
            f'{samples}/dumped'
            f'/{planet}.json')

        source = rvrt_sample(
            sample=source,
            replace=replaces)

        sample = loads(source)

        sitem = (
            origin.get_stream(
                sample[3]['event']))

        origin.set_update(sitem)

        fetch = origin.fetch


        assert _fetch != fetch


    assert uqueue.qsize == 2



def test_PhueOrigin_stream(
    homie: 'Homie',
    service: 'HomieService',
    bodies: 'TestBodies',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    :param service: Ancilary Homie Automate class instance.
    :param bodies: Locations and groups for use in testing.
    """

    childs = homie.childs
    origins = childs.origins
    member = service.streams
    squeue = member.squeue


    planets = bodies.planets

    event = {'foo': 'bar'}

    for planet in planets:

        origin = origins[
            f'{planet}_philips']

        assert isinstance(
            origin, PhueOrigin)

        origin.put_stream(
            squeue, event)


    assert squeue.qsize == 2



def test_PhueOrigin_samples(
    homie: 'Homie',
    replaces: DictStrAny,
    bodies: 'TestBodies',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    :param replaces: Mapping of what to replace in samples.
    :param bodies: Locations and groups for use in testing.
    """

    childs = homie.childs
    origins = childs.origins

    samples = SAMPLES / 'origin'


    planets = bodies.planets

    for planet in planets:

        origin = origins[
            f'{planet}_philips']

        assert isinstance(
            origin, PhueOrigin)

        assert origin.refresh()


        sample_path = (
            f'{samples}/fetch'
            f'/{planet}.json')

        sample = load_sample(
            path=sample_path,
            update=ENPYRWS,
            content=origin.fetch,
            replace=replaces)

        expect = prep_sample(
            content=origin.fetch,
            replace=replaces)

        assert expect == sample


        sample_path = (
            f'{samples}/merge'
            f'/{planet}.json')

        sample = load_sample(
            path=sample_path,
            update=ENPYRWS,
            content=origin.merge,
            replace=replaces)

        expect = prep_sample(
            content=origin.merge,
            replace=replaces)

        assert expect == sample



def test_PhueOrigin_source(
    homie: 'Homie',
    bodies: 'TestBodies',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    :param bodies: Locations and groups for use in testing.
    """

    childs = homie.childs
    origins = childs.origins
    groups = childs.groups


    planets = bodies.planets
    moons = dict(bodies.moons)

    for planet in planets:

        origin = origins[
            f'{planet}_philips']

        origin.refresh()


        source = origin.source(
            label='Nothing')

        assert source is None


        moon = moons[planet]

        source = origin.source(
            kind='scene',
            label='active',
            relate=groups[moon])

        assert source is not None


        source = origin.source(
            kind='group',
            label=planet)

        assert source is not None

        source = origin.source(
            kind='group',
            label=moon)

        assert source is not None



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

from encommon.times import Times

from ...utils import UnexpectedCondition

if TYPE_CHECKING:
    from ..helpers import UbiqFetch



def ubiq_latest(
    source: 'UbiqFetch',
) -> Times:
    """
    Return the timestamp for client association with router.

    :param source: Content which will be shown after header.
    :returns: Timestamp for client association with router.
    """


    gwsecs = (
        source
        .get('_uptime_by_ugw'))

    apsecs = (
        source
        .get('_uptime_by_uap'))

    if (gwsecs is not None
            or apsecs is not None):
        return Times()


    laseen = (
        source
        .get('last_seen'))

    if laseen is not None:
        return Times(laseen[0])


    raise UnexpectedCondition



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



class UnexpectedCondition(Exception):
    """
    Exception when action was not taken due to idempotency.
    """



class Idempotent(Exception):
    """
    Exception when action was not taken due to idempotency.
    """



class MultipleSource(Exception):
    """
    Exception when action was not taken due to idempotency.
    """



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

from ..update import UbiqUpdateItem

if TYPE_CHECKING:
    from ...homie import HomieService
    from ...utils import TestBodies



def test_UbiqUpdateItem(
    service: 'HomieService',
    bodies: 'TestBodies',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param service: Ancilary Homie Automate class instance.
    :param bodies: Locations and groups for use in testing.
    """

    homie = service.homie
    childs = homie.childs
    origins = childs.origins

    model = UbiqUpdateItem


    planets = bodies.planets

    fetch = {'foo': 'bar'}

    for planet in planets:

        origin = origins[
            f'{planet}_ubiquiti']


        item = model(
            origin, fetch)


        assert item.origin == origin.name

        assert item.fetch == fetch



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from argparse import ArgumentParser
from sys import argv as sys_argv
from typing import Optional
from typing import TYPE_CHECKING
from typing import get_args

from encommon.types import DictStrAny

from ..homie import Homie
from ..homie import HomieConfig
from ..homie.common import HomieState

if TYPE_CHECKING:
    from ..homie.childs import HomieDevice
    from ..homie.childs import HomieGroup
    from ..homie.childs import HomieScene



def arguments(  # noqa: CFQ001
    args: Optional[list[str]] = None,
) -> DictStrAny:
    """
    Construct arguments which are associated with the file.

    :param args: Override the source for the main arguments.
    :returns: Construct arguments from command line options.
    """

    parser = ArgumentParser()

    args = args or sys_argv[1:]


    parser.add_argument(
        '--config',
        required=True,
        help=(
            'complete or relative '
            'path to config file'))


    parser.add_argument(
        '--console',
        action='store_true',
        default=False,
        help=(
            'write log messages '
            'to standard output'))


    parser.add_argument(
        '--debug',
        action='store_true',
        default=False,
        help=(
            'increase logging level '
            'for standard output'))


    parser.add_argument(
        '--dry-run',
        action='store_true',
        default=False,
        dest='dryrun',
        help='do not execute actions')


    parser.add_argument(
        '--idempotent',
        action='store_false',
        default=None,
        dest='potent',
        help=(
            'do not make requests '
            'when already applied'))


    parser.add_argument(
        '--print',
        action='store_true',
        default=False,
        dest='paction',
        help=(
            'print the submited '
            'actions to console'))


    parser.add_argument(
        '--group',
        help=(
            'which group action '
            'is performed upon'))


    parser.add_argument(
        '--device',
        help=(
            'which device action '
            'is performed upon'))


    states = list(
        get_args(HomieState))

    parser.add_argument(
        '--state',
        choices=states,
        help=(
            'which state value the '
            'target will changed'))


    parser.add_argument(
        '--color',
        help=(
            'which color value the '
            'target will changed'))


    parser.add_argument(
        '--level',
        type=int,
        help=(
            'which level value the '
            'target will changed'))


    parser.add_argument(
        '--scene',
        help=(
            'which level value the '
            'target will changed'))


    return vars(
        parser
        .parse_args(args))



def operation(
    # NOCVR
    homie: Homie,
) -> None:
    """
    Perform whatever operation is associated with the file.

    :param homie: Primary class instance for Homie Automate.
    """

    assert homie.refresh()

    config = homie.config
    sargs = config.sargs

    _group = sargs['group']
    _device = sargs['device']
    _scene = sargs['scene']

    childs = homie.childs
    devices = childs.devices
    groups = childs.groups
    scenes = childs.scenes
    potent = homie.potent


    group: Optional['HomieGroup']
    device: Optional['HomieDevice']
    scene: Optional['HomieScene']

    group = (
        groups[_group]
        if _group is not None
        else None)

    device = (
        devices[_device]
        if _device is not None
        else None)

    scene = (
        scenes[_scene]
        if _scene is not None
        else None)


    target = device or group

    assert target is not None


    aitems = (
        homie.get_actions(
            target=target,
            state=sargs['state'],
            color=sargs['color'],
            level=sargs['level'],
            scene=scene))

    homie.set_actions(
        aitems=aitems,
        force=potent)



def execution(
    # NOCVR
    args: Optional[list[str]] = None,
) -> None:
    """
    Perform whatever operation is associated with the file.

    :param args: Override the source for the main arguments.
    """

    config = HomieConfig(
        arguments(args))

    config.logger.start()

    homie = Homie(config)

    config.logger.log_i(
        base='execution/action',
        status='started')

    operation(homie)

    config.logger.log_i(
        base='execution/action',
        status='stopped')

    config.logger.stop()



if __name__ == '__main__':
    execution()  # NOCVR



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from contextlib import suppress
from typing import Optional

from encommon.config import Config
from encommon.config import Params
from encommon.types import DictStrAny
from encommon.utils.common import PATHABLE

from .params import HomieParams



class HomieConfig(Config):
    """
    Contain the configurations from the arguments and files.

    :param sargs: Additional arguments on the command line.
    :param files: Complete or relative path to config files.
    :param cargs: Configuration arguments in dictionary form,
        which will override contents from the config files.
    """


    def __init__(
        self,
        sargs: Optional[DictStrAny] = None,
        files: Optional[PATHABLE] = None,
        cargs: Optional[DictStrAny] = None,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        sargs = dict(sargs or {})
        cargs = dict(cargs or {})


        _dryrun = sargs.get('dryrun')
        _potent = sargs.get('potent')

        if _dryrun is not None:
            cargs['dryrun'] = _dryrun

        if _potent is not None:
            cargs['potent'] = _potent


        _idesire = sargs.get('idesire')
        _iupdate = sargs.get('iupdate')
        _ihealth = sargs.get('ihealth')

        prefix = 'service/respite'

        if _idesire is not None:
            key = f'{prefix}/desire'
            cargs[key] = _idesire

        if _iupdate is not None:
            key = f'{prefix}/update'
            cargs[key] = _iupdate

        if _ihealth is not None:
            key = f'{prefix}/health'
            cargs[key] = _ihealth


        _atimeout = sargs.get('atimeout')
        _utimeout = sargs.get('utimeout')
        _stimeout = sargs.get('stimeout')

        prefix = 'service/timeout'

        if _atimeout is not None:
            key = f'{prefix}/action'
            cargs[key] = _atimeout

        if _utimeout is not None:
            key = f'{prefix}/update'
            cargs[key] = _utimeout

        if _stimeout is not None:
            key = f'{prefix}/stream'
            cargs[key] = _stimeout

        _paction = sargs.get('paction')
        _pupdate = sargs.get('pupdate')
        _pstream = sargs.get('pstream')
        _pdesire = sargs.get('pdesire')
        _paspire = sargs.get('paspire')

        if _paction is not None:
            key = 'printer/action'
            cargs[key] = _paction

        if _pupdate is not None:
            key = 'printer/update'
            cargs[key] = _pupdate

        if _pstream is not None:
            key = 'printer/stream'
            cargs[key] = _pstream

        if _pdesire is not None:
            key = 'printer/desire'
            cargs[key] = _pdesire

        if _paspire is not None:
            key = 'printer/aspire'
            cargs[key] = _paspire


        _debug = sargs.get('debug')
        _console = sargs.get('console')

        key = 'enlogger/stdo_level'

        if _console is True:
            cargs[key] = 'info'

        if _debug is True:
            cargs[key] = 'debug'


        if 'config' in sargs:
            files = sargs['config']


        super().__init__(
            files=files,
            cargs=cargs,
            sargs=sargs,
            model=HomieParams)


    @property
    def params(
        self,
    ) -> HomieParams:
        """
        Return the Pydantic model containing the configuration.

        .. warning::
           This method completely overrides the parent but is
           based on that code, would be unfortunate if upstream
           changes meant this breaks or breaks something else.

        :returns: Pydantic model containing the configuration.
        """

        params = self.__params

        if params is not None:

            assert isinstance(
                params,
                HomieParams)

            return params


        basic = self.basic
        update = False

        with suppress(AttributeError):
            basic = self.merge
            update = True


        params = (
            self.model(**basic))

        assert isinstance(
            params,
            HomieParams)


        if update is True:
            self.__params = params

        return params


    @property
    def __params(
        self,
    ) -> Optional[Params]:
        """
        Return the Pydantic model containing the configuration.

        :returns: Pydantic model containing the configuration.
        """

        return self._Config__params


    @__params.setter
    def __params(
        self,
        value: Params,
    ) -> None:
        """
        Update the value for the attribute from class instance.
        """

        self._Config__params = value



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from .child import InvalidChild
from .param import InvalidParam
from .raises import Idempotent
from .raises import MultipleSource
from .raises import UnexpectedCondition
from .tests import TestBodies
from .tests import TestTimes



__all__ = [
    'InvalidChild',
    'InvalidParam',
    'UnexpectedCondition',
    'Idempotent',
    'MultipleSource',
    'TestTimes',
    'TestBodies']



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Annotated
from typing import Any
from typing import Optional

from encommon.types import BaseModel

from pydantic import Field

from ...plugins.button import PhueButtonEvent
from ...plugins.button import PhueButtonSensor



class DriverPhueButtonParams(BaseModel, extra='forbid'):
    """
    Process and validate the Homie configuration parameters.
    """

    device: Annotated[
        str,
        Field(...,
              description='Device that is in scope',
              min_length=1)]

    events: Annotated[
        Optional[list[PhueButtonEvent]],
        Field(None,
              description='Events that will be matched',
              min_length=1)]

    sensor: Annotated[
        PhueButtonSensor,
        Field(...,
              description='Sensor that will be matched')]


    def __init__(
        # NOCVR
        self,
        **data: Any,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        events = data.get('events')

        if isinstance(events, str):
            data['events'] = [events]

        super().__init__(**data)



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Literal
from typing import Optional
from typing import TYPE_CHECKING

from encommon.times import Times
from encommon.times import findtz

from .models import BltnModels
from ..homie.plugins import HomieDriver
from ..utils import InvalidParam

if TYPE_CHECKING:
    from .params import DriverBltnPeriodParams
    from ..homie.plugins import HomiePluginKinds



class DriverBltnPeriod(HomieDriver):
    """
    Match specific conditions for determining desired state.
    """


    def validate(
        self,
    ) -> None:
        """
        Perform advanced validation on the parameters provided.
        """

        params = self.params

        _start = params.start
        _stop = params.stop
        _tzname = params.tzname


        if (_start and _stop
                and _start >= _stop):

            raise InvalidParam(
                param='start',
                value=_start,
                error='invalid')


        tzinfo = findtz(_tzname)

        if tzinfo is None:

            raise InvalidParam(
                param='tzname',
                value=_tzname,
                error='invalid')


    @property
    def family(
        self,
    ) -> Literal['builtins']:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return 'builtins'


    @property
    def kinds(
        self,
    ) -> list['HomiePluginKinds']:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return ['where']


    @property
    def params(
        self,
    ) -> 'DriverBltnPeriodParams':
        """
        Return the Pydantic model containing the configuration.

        :returns: Pydantic model containing the configuration.
        """

        model = (
            BltnModels
            .drivers()
            .period())

        params = super().params

        assert isinstance(
            params, model)

        return params


    def where(  # noqa: CFQ004
        self,
        time: Times,
    ) -> bool:
        """
        Return the boolean indicating the conditional outcomes.

        :param time: Time that will be used in the conditionals.
        :returns: Boolean indicating the conditional outcomes.
        """

        params = self.params
        tzname = params.tzname

        start: Optional[Times] = None
        stop: Optional[Times] = None

        if tzname is not None:
            time = time.shifz(tzname)


        _time = Times(
            time.stamp('%H:%M'))

        today = time.stamp('%A')


        if params.start:

            start = Times(
                params.start,
                tzname=tzname)

            _start = Times(
                start.stamp('%H:%M'))

            if _time < _start:
                return False

        if params.stop:

            stop = Times(
                params.stop,
                tzname=tzname)

            _stop = Times(
                stop.stamp('%H:%M'))

            if _time > _stop:
                return False


        if params.days:

            days = params.days

            if today not in days:
                return False


        return True



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from dataclasses import dataclass
from typing import Literal
from typing import Optional
from typing import TYPE_CHECKING
from typing import Union

from encommon.colors import Colors
from encommon.utils.stdout import ANSIARRAY

if TYPE_CHECKING:
    from .threads import HomieThreadItem
    from .addons import HomieAspiredItem
    from .addons import HomieDesiredItem



HomieFamily = Literal[
    'builtins',
    'hubitat',
    'philips',
    'ubiquiti']

HomieState = Literal[
    'poweron',
    'nopower']

HomiePrint = Union[
    ANSIARRAY,
    'HomieThreadItem',
    'HomieDesiredItem',
    'HomieAspiredItem']



@dataclass
class HomieStage:
    """
    Contain information about the parameters for the scene.
    """

    state: Optional[HomieState] = None
    color: Optional[Colors] = None
    level: Optional[int] = None


    def __init__(
        self,
        state: Optional[HomieState] = None,
        color: Optional[str | Colors] = None,
        level: Optional[int] = None,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        if isinstance(color, str):
            color = Colors(color)

        self.state = state
        self.color = color
        self.level = level



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

from encommon.types import inlist
from encommon.types import inrepr
from encommon.types import instr
from encommon.types import lattrs

from ..action import HomieActionItem

if TYPE_CHECKING:
    from ...service import HomieService
    from ....utils import TestBodies



def test_HomieActionItem(
    service: 'HomieService',
    bodies: 'TestBodies',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param service: Ancilary Homie Automate class instance.
    :param bodies: Locations and groups for use in testing.
    """

    homie = service.homie
    childs = homie.childs
    origins = childs.origins
    devices = childs.devices
    groups = childs.groups
    scenes = childs.scenes

    model = HomieActionItem


    moons = (
        dict(bodies.moons)
        .items())

    for planet, moon in moons:

        origin = origins[
            f'{planet}_philips']

        device = devices[
            f'{planet}_light1']

        group = groups[moon]
        scene = scenes['active']


        item = model(
            origin, group,
            color='ff00cc')


        attrs = lattrs(item)

        assert attrs == [
            'group',
            'state',
            'color',
            'level',
            'scene',
            'origin',
            'time']


        assert item.origin == origin.name

        assert item.group == group.name

        assert not item.device

        assert not item.state

        assert item.color == 'ff00cc'

        assert not item.level


        item = model(
            origin, device,
            scene=scene)

        assert not item.group

        assert item.device == device.name

        assert item.scene == scene.name



def test_HomieAction(
    service: 'HomieService',
    bodies: 'TestBodies',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param service: Ancilary Homie Automate class instance.
    :param bodies: Locations and groups for use in testing.
    """

    member = service.actions
    threads = member.threads


    planets = bodies.planets

    for planet in planets:

        thread = threads[
            f'{planet}_philips']


        attrs = lattrs(thread)

        # Family class inheritence

        assert inlist(
            '_HomieThread__member',
            attrs)

        assert inlist(
            '_HomieThread__origin',
            attrs)


        assert inrepr(
            'Action(PhueAction',
            thread)

        assert hash(thread) > 0

        assert instr(
            'Action(PhueAction',
            thread)


        assert thread.homie

        assert thread.member

        assert thread.origin

        assert thread.aqueue

        assert thread.uqueue

        assert thread.squeue



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from dataclasses import dataclass
from typing import Any
from typing import Optional

from ..homie.threads import HomieAction
from ..homie.threads import HomieActionItem
from ..homie.threads import HomieActionNode



@dataclass
class HubiActionItem(HomieActionItem):
    """
    Contain information for sharing using the Python queue.
    """


    def __init__(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        super().__init__(
            *args, **kwargs)



class HubiAction(HomieAction):
    """
    Common methods and routines for Homie Automate threads.
    """


    def execute(
        self,
        aitem: HomieActionItem,
    ) -> None:
        """
        Perform the operation related to Homie service threads.

        :param aitem: Item containing information for operation.
        """

        assert isinstance(
            aitem, HubiActionItem)

        homie = self.homie
        childs = homie.childs
        devices = childs.devices

        params = homie.params
        potent = homie.potent

        timeout = (
            params.service
            .timeout.action)

        _device = aitem.device
        _group = aitem.group


        target: Optional[
            HomieActionNode]

        target = None

        if _device is not None:
            target = devices[_device]

        assert _group is None

        if target is None:

            homie.logger.log_e(
                base=self,
                item=aitem,
                status='notarget')

            return None


        origin = target.origin

        assert origin is not None

        origin.set_action(
            target=target,
            aitem=aitem,
            force=potent,
            timeout=timeout)



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Annotated
from typing import Any

from encommon.times import unitime
from encommon.types import BaseModel

from pydantic import Field

from ..addons import HomiePersistExpire
from ..addons import HomiePersistValue



class HomieStoreParams(BaseModel, extra='forbid'):
    """
    Process and validate the Homie configuration parameters.
    """

    unique: Annotated[
        str,
        Field(...,
              description='Unique key for the value',
              min_length=1)]

    value: Annotated[
        HomiePersistValue,
        Field(...,
              description='Value stored at the key')]

    expire: Annotated[
        HomiePersistExpire,
        Field('1d',
              description='After when the key expires')]


    def __init__(
        # NOCVR
        self,
        **data: Any,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        expire = data.get('expire')

        if isinstance(expire, str):
            expire = unitime(expire)
            assert expire >= 0

        super().__init__(**data)



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from .common import HomiePluginKinds
from .driver import HomieDriver
from .occur import HomieOccur
from .plugin import HomiePlugin
from .where import HomieWhere



__all__ = [
    'HomiePlugin',
    'HomiePluginKinds',
    'HomieWhere',
    'HomieOccur',
    'HomieDriver']



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from .button import DriverPhueButton
from .change import DriverPhueChange
from .contact import DriverPhueContact
from .motion import DriverPhueMotion
from .scene import DriverPhueScene



__all__ = [
    'DriverPhueButton',
    'DriverPhueChange',
    'DriverPhueContact',
    'DriverPhueMotion',
    'DriverPhueScene']



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Optional
from typing import TYPE_CHECKING
from typing import Type

from encommon.times import Timer
from encommon.times import Times

from .member import HomieMember
from ...hubitat import HubiAction
from ...philips import PhueAction

if TYPE_CHECKING:
    from ..common import HomieState
    from ..childs import HomieOrigin
    from ..childs import HomieScene
    from ..threads import HomieActionNode
    from ..threads import HomieThread



class HomieActions(HomieMember):
    """
    Common methods and routines for Homie Automate members.
    """

    __timer: Timer


    def __post__(
        self,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        homie = self.homie
        params = homie.params

        respite = (
            params.service
            .respite)

        timer = Timer(
            respite.desire,
            start='min')

        self.__timer = timer


    def operate(
        self,
    ) -> None:
        """
        Perform the operation related to Homie service members.
        """

        homie = self.homie

        self.operate_updates()


        try:
            self.operate_desires()
            self.operate_actions()

        except Exception as reason:

            homie.logger.log_e(
                base=self,
                item='desired',
                status='failure',
                exc_info=reason)


        try:
            self.operate_aspires()
            self.operate_actions()

        except Exception as reason:

            homie.logger.log_e(
                base=self,
                item='aspired',
                status='failure',
                exc_info=reason)


    def operate_desires(
        self,
    ) -> None:
        """
        Perform the operation related to Homie service members.
        """

        homie = self.homie
        desired = homie.desired
        timer = self.__timer

        if not timer.ready():
            return None

        vacate = self.vacate

        if vacate.is_set():
            return None


        def _execute() -> None:

            homie.printer(aitem)

            self.put_actions(
                aitem.target,
                state=aitem.state,
                color=aitem.color,
                level=aitem.level,
                scene=aitem.scene)


        homie.logger.log_d(
            base=self,
            item='desired',
            status='started')

        time = Times('now')

        aitems = (
            desired.items(time))

        for aitem in aitems:
            _execute()

        homie.logger.log_d(
            base=self,
            item='desired',
            count=len(aitems),
            status='complete',
            elapsed=time)


    def operate_aspires(
        self,
    ) -> None:
        """
        Perform the operation related to Homie service members.
        """

        homie = self.homie
        aspired = homie.aspired

        vacate = self.vacate
        squeue = self.squeue


        def _execute() -> None:

            homie.printer(aitem)

            self.put_actions(
                aitem.target,
                state=aitem.state,
                color=aitem.color,
                level=aitem.level,
                scene=aitem.scene)


        while not squeue.empty:

            sitem = squeue.get()

            if vacate.is_set():
                continue

            aitems = (
                aspired.items(sitem))

            for aitem in aitems:
                _execute()


    def operate_actions(
        self,
    ) -> None:
        """
        Perform the operation related to Homie service members.
        """

        # Where the actions process

        homie = self.homie
        childs = homie.childs
        origins = childs.origins
        vacate = self.vacate
        aqueue = self.aqueue

        threads = (
            self.threads
            .values())


        def _put_action() -> None:

            origin = thread.origin
            _name = origin.name

            if name != _name:
                return None

            (thread
             .aqueue.put(aitem))


        while not aqueue.empty:

            aitem = aqueue.get()

            if vacate.is_set():
                continue

            origin = origins[
                aitem.origin]

            name = origin.name

            for thread in threads:
                _put_action()

            homie.printer(aitem)


    def get_thread(
        self,
        origin: 'HomieOrigin',
    ) -> Optional[Type['HomieThread']]:
        """
        Return the Homie class definition for its instantiation.

        :param origin: Child class instance for Homie Automate.
        :returns: Homie class definition for its instantiation.
        """

        family = origin.family

        if family == 'hubitat':
            return HubiAction

        if family == 'philips':
            return PhueAction

        return None


    def put_actions(
        self,
        target: 'HomieActionNode',
        *,
        state: Optional['HomieState'] = None,
        color: Optional[str] = None,
        level: Optional[int] = None,
        scene: Optional['HomieScene'] = None,
    ) -> None:
        """
        Insert the new item containing information for operation.

        :param target: Device or group settings will be updated.
        :param state: Determine the state related to the target.
        :param color: Determine the color related to the target.
        :param level: Determine the level related to the target.
        :param scene: Determine the scene related to the target.
        """

        homie = self.homie
        aqueue = self.aqueue

        aitems = (
            homie.get_actions(
                target=target,
                state=state,
                color=color,
                level=level,
                scene=scene))

        items = aitems.items()

        for target, aitem in items:

            origin = target.origin

            assert origin is not None

            origin.put_action(
                aqueue=aqueue,
                target=target,
                state=aitem.state,
                color=aitem.color,
                level=aitem.level,
                scene=aitem.scene)



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ...service import HomieService



def test_HomieMember_cover(
    service: 'HomieService',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param service: Ancilary Homie Automate class instance.
    """

    # Remove if not needed



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Annotated
from typing import Literal
from typing import Optional

from encommon.types import BaseModel

from pydantic import Field

from ..store import BltnStoreOperas
from ...homie.addons import HomiePersistValue



class DriverBltnStoreParams(BaseModel, extra='forbid'):
    """
    Process and validate the Homie configuration parameters.
    """

    unique: Annotated[
        str,
        Field(...,
              description='Unique key for the value',
              min_length=1)]

    operator: Annotated[
        Literal[BltnStoreOperas],
        Field(...,
              description='Operator for the condition')]

    value: Annotated[
        Optional[HomiePersistValue],
        Field(None,
              description='Value for the condition')]



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from ..desired import arguments



def test_arguments() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    sargs = arguments([
        '--config', 'path'])

    assert sargs == {
        'config': 'path',
        'console': False,
        'debug': False,
        'dryrun': True,
        'pdesire': True}



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Annotated
from typing import Any
from typing import Optional

from pydantic import Field

from .child import HomieChildParams



class HomieGroupParams(HomieChildParams, extra='forbid'):
    """
    Process and validate the Homie configuration parameters.
    """

    origin: Annotated[
        Optional[str],
        Field(None,
              description='Origin where group exists',
              min_length=1)]

    label: Annotated[
        Optional[str],
        Field(None,
              description='Group name in the origin',
              min_length=1)]

    devices: Annotated[
        Optional[list[str]],
        Field(None,
              description='List of devices in group',
              min_length=1)]


    def __init__(
        # NOCVR
        self,
        **data: Any,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        devices = data.get('devices')

        if isinstance(devices, str):
            data['devices'] = [devices]

        super().__init__(**data)



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from threading import Thread
from time import sleep as block_sleep
from typing import TYPE_CHECKING

from encommon.types import inrepr
from encommon.types import instr
from encommon.types import lattrs

if TYPE_CHECKING:
    from ..service import HomieService



def test_HomieService(
    service: 'HomieService',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param service: Ancilary Homie Automate class instance.
    """


    attrs = lattrs(service)

    assert attrs == [
        '_HomieService__homie',
        '_HomieService__actions',
        '_HomieService__updates',
        '_HomieService__streams',
        '_HomieService__timer',
        '_HomieService__vacate',
        '_HomieService__cancel',
        '_HomieService__started']


    assert inrepr(
        'service.HomieService',
        service)

    assert hash(service) > 0

    assert instr(
        'service.HomieService',
        service)


    assert service.homie

    assert service.params

    assert service.actions

    assert service.updates

    assert service.streams

    assert len(service.running) == 0

    assert len(service.zombies) == 12


    service.start()

    operate = service.operate

    thread = Thread(
        target=operate)

    thread.start()

    for _ in range(50):
        thread.join(0.1)

    assert service.running

    assert not service.zombies

    assert service.congest

    service.soft()

    while service.congest:
        thread.join(0.1)

    while service.running:
        thread.join(0.1)

    service.stop()

    thread.join()

    assert service.zombies

    assert not service.congest



def test_HomieService_dryrun(
    service: 'HomieService',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param service: Ancilary Homie Automate class instance.
    """

    homie = service.homie
    params = homie.params

    params.dryrun = True


    service.start()

    operate = service.operate

    thread = Thread(
        target=operate)

    thread.start()

    for _ in range(50):
        thread.join(0.1)

    assert service.running

    assert not service.zombies

    assert service.congest

    service.soft()

    while service.congest:
        thread.join(0.1)

    while service.running:
        thread.join(0.1)

    service.stop()

    thread.join()

    assert service.zombies

    assert not service.congest


    block_sleep(1.1)

    service.operate_healths()



def test_HomieService_cover(
    service: 'HomieService',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param service: Ancilary Homie Automate class instance.
    """

    service.stop()
    service.soft()
    service.start()
    service.start()
    service.soft()
    service.soft()
    service.stop()
    service.stop()



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

from encommon.types import DictStrAny
from encommon.types import inrepr
from encommon.types import instr
from encommon.types import lattrs
from encommon.utils import load_sample
from encommon.utils import prep_sample
from encommon.utils.sample import ENPYRWS

from . import SAMPLES
from ..config import HomieConfig
from ..homie import Homie
from ..threads import HomieThreadItem

if TYPE_CHECKING:
    from ...utils import TestBodies



def test_Homie(
    homie: Homie,
    replaces: DictStrAny,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    :param replaces: Mapping of what to replace in samples.
    """


    attrs = lattrs(homie)

    assert attrs == [
        '_Homie__config',
        '_Homie__logger',
        '_Homie__persist',
        '_Homie__childs',
        '_Homie__desired',
        '_Homie__aspired']


    assert inrepr(
        'homie.Homie',
        homie)

    assert hash(homie) > 0

    assert instr(
        'homie.Homie',
        homie)


    assert homie.config

    assert homie.logger

    assert homie.persist

    assert homie.childs

    assert homie.params

    assert homie.desired

    assert homie.aspired


    assert homie.refresh()


    sample_path = (
        SAMPLES / 'dumped.json')

    sample = load_sample(
        path=sample_path,
        update=ENPYRWS,
        content=homie.dumped,
        replace=replaces)

    expect = prep_sample(
        content=homie.dumped,
        replace=replaces)

    assert expect == sample



def test_Homie_actions(
    homie: Homie,
    bodies: 'TestBodies',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    :param bodies: Locations and groups for use in testing.
    """

    childs = homie.childs
    devices = childs.devices
    groups = childs.groups


    assert homie.refresh()


    moons = (
        dict(bodies.moons)
        .items())

    for planet, moon in moons:


        device = devices[
            f'{planet}_light2']

        aitems = (
            homie.get_actions(
                target=device,
                state='poweron',
                color='ff00cc',
                level=69))

        assert len(aitems) == 1

        assert homie.set_actions(
            aitems, force=True)

        assert homie.set_actions(
            aitems, change=False)


        device = devices[
            f'{planet}_light1']

        aitems = (
            homie.get_actions(
                target=device,
                state='nopower',
                color='ff00cc',
                level=100))

        assert len(aitems) == 1

        assert homie.set_actions(
            aitems, force=True)

        assert homie.set_actions(
            aitems, change=False)


        group = groups[moon]

        aitems = (
            homie.get_actions(
                target=group,
                state='poweron',
                color='ff00cc',
                level=69))

        assert len(aitems) == 2

        assert homie.set_actions(
            aitems, force=True)

        assert homie.set_actions(
            aitems, change=False)


        aitems = (
            homie
            .get_actions(group))

        assert len(aitems) == 0



def test_Homie_printer(
    homie: Homie,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    """

    childs = homie.childs
    origins = childs.origins

    origin = origins[
        'jupiter_philips']

    model = HomieThreadItem

    item = model(origin)

    homie.printer(item)



def test_Homie_cover(
) -> None:
    """
    Perform various tests associated with relevant routines.
    """

    Homie(HomieConfig())



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from copy import deepcopy
from json import dumps
from typing import Literal
from typing import Optional
from typing import TYPE_CHECKING
from urllib.parse import quote_plus

from encommon.colors import Colors
from encommon.types import DictStrAny
from encommon.types import getate
from encommon.types import strplwr

from httpx import Response

from .device import HubiDevice
from ..utils import Idempotent
from ..utils import UnexpectedCondition

if TYPE_CHECKING:
    from .origin import HubiOrigin
    from ..homie.childs import HomieChildKinds
    from ..homie.common import HomieState
    from ..homie.childs import HomieScene
    from ..homie.threads import HomieActionNode



HubiFetch = list[DictStrAny]
HubiMerge = dict[str, DictStrAny]



_RESPONSE = Optional[
    Response | Literal[True]]

_APROPS = Literal[
    'state',
    'color',
    'level']

_ASTATES = Literal[
    'on', 'off']

_AVALUES = Optional[
    _ASTATES
    | int | Colors]



def merge_fetch(
    fetch: HubiFetch,
) -> HubiMerge:
    """
    Return the content related to the item in parent system.

    :param fetch: Dictionary from the origin to be updated.
    :returns: Content related to the item in parent system.
    """

    fetch = deepcopy(fetch)

    source = {
        x['id']: x
        for x in fetch}

    _source = sorted(
        source.items())

    return dict(_source)



def merge_find(  # noqa: CFQ004
    merge: HubiMerge,
    kind: Optional['HomieChildKinds'] = None,
    unique: Optional[str] = None,
    label: Optional[str] = None,
) -> list[DictStrAny]:
    """
    Return the content related to the item in parent system.

    :param merge: Dictionary from the origin to be updated.
    :param kind: Which kind of Homie object will be located.
    :param unique: Unique identifier within parents system.
    :param label: Friendly name or label within the parent.
    :returns: Content related to the item in parent system.
    """

    assert unique or label

    if label is not None:
        label = strplwr(label)

    if unique is not None:
        unique = strplwr(unique)

    found: list[DictStrAny] = []


    def _match_kind() -> bool:

        if kind is None:
            return True

        _kind: Optional[str] = None

        _capas = value[
            'capabilities']

        if 'Light' in _capas:
            _kind = 'device'

        return kind == _kind


    def _match_unique() -> bool:

        if unique is None:
            return True

        _unique = strplwr(
            value['id'])

        return unique == _unique


    def _match_label() -> bool:

        if label is None:
            return True

        _label = strplwr(
            value['label'])

        return label == _label


    values = merge.values()

    for value in values:

        if not _match_kind():
            continue

        if not _match_unique():
            continue

        if not _match_label():
            continue

        found.append(value)


    return found



def request_action(  # noqa: CFQ001,CFQ002,CFQ004
    origin: 'HubiOrigin',
    target: 'HomieActionNode',
    *,
    state: Optional['HomieState'] = None,
    color: Optional[str | Colors] = None,
    level: Optional[int] = None,
    scene: Optional['HomieScene'] = None,
    force: bool = False,
    change: bool = True,
    timeout: Optional[int] = None,
) -> None:
    """
    Perform the provided action with specified Homie target.

    :param origin: Child class instance for Homie Automate.
    :param target: Device or group settings will be updated.
    :param state: Determine the state related to the target.
    :param color: Determine the color related to the target.
    :param level: Determine the level related to the target.
    :param scene: Determine the scene related to the target.
    :param force: Override the default for full idempotency.
    :param change: Determine whether the change is executed.
    :param timeout: Timeout waiting for the server response.
    """

    homie = origin.homie
    kind = target.kind

    request = action_request

    assert kind == 'device'

    changed: set[bool] = set()


    source = target.source


    if scene is not None:

        assert isinstance(
            target, HubiDevice)

        stage = (
            scene
            .stage(target))

        if state is None:
            state = stage.state

        if color is None:
            color = stage.color

        if level is None:
            level = stage.level


    assert source is not None

    unique = source['id']


    _state: _AVALUES = (
        'off'
        if state == 'nopower'
        else 'on')


    if isinstance(color, str):
        color = Colors(color)


    def _set_state() -> None:

        response = request(
            origin=origin,
            unique=unique,
            property='state',
            value=_state,
            force=force,
            change=change,
            timeout=timeout)

        if response is None:
            return None

        changed.add(True)

        if (change is False
                or homie.dryrun):
            return None

        assert isinstance(
            response, Response)

        (response
         .raise_for_status())


    def _set_color() -> None:

        response = request(
            origin=origin,
            unique=unique,
            property='color',
            value=color,
            force=force,
            change=change,
            timeout=timeout)

        if response is None:
            return None

        changed.add(True)

        if (change is False
                or homie.dryrun):
            return None

        assert isinstance(
            response, Response)

        (response
         .raise_for_status())


    def _set_level() -> None:

        response = request(
            origin=origin,
            unique=unique,
            property='level',
            value=level,
            force=force,
            change=change,
            timeout=timeout)

        if response is None:
            return None

        changed.add(True)

        if (change is False
                or homie.dryrun):
            return None

        assert isinstance(
            response, Response)

        (response
         .raise_for_status())


    if state is not None:
        _set_state()

    if state != 'nopower':

        if color is not None:
            _set_color()

        if level is not None:
            _set_level()


    if not any(changed):
        raise Idempotent



def action_request(  # noqa: CFQ001,CFQ002,CFQ004
    origin: 'HubiOrigin',
    unique: str,
    *,
    property: _APROPS,
    value: _AVALUES = None,
    force: bool = False,
    change: bool = True,
    timeout: Optional[int] = None,
) -> _RESPONSE:
    """
    Request to execute the action on target from the bridge.

    :param origin: Child class instance for Homie Automate.
    :param type: Which type of target to perform the action.
    :param unique: Unique identifier within parents system.
    :param property: What attribute on the target is change.
    :param value: Value relating to property being updated.
    :param force: Override the default for full idempotency.
    :param change: Determine whether the change is executed.
    :param timeout: Timeout waiting for the server response.
    :returns: Response from upstream request to the server.
    """

    bridge = origin.bridge
    request = bridge.request
    homie = origin.homie

    merge = origin.source(
        unique=unique)

    assert merge is not None


    def _get_attr(
        attr: str,
    ) -> str:

        pre = 'attributes'

        key = (
            'saturation'
            if attr == 'sat'
            else attr)

        value = getate(
            merge,
            f'{pre}/{key}')

        return str(value)


    def _set_state() -> _RESPONSE:

        assert isinstance(
            value, str)

        crrnt = _get_attr('switch')

        path = (
            f'devices/{unique}'
            f'/{value}')

        if (crrnt == value
                and not force):
            return None

        if (change is False
                or homie.dryrun):
            return True

        return request(
            method='get',
            path=path,
            timeout=timeout)


    def _set_color() -> _RESPONSE:

        assert isinstance(
            value, Colors)

        hsl = list(value.hsl)

        hsb = (
            int(hsl[0] / 3.6),
            int(hsl[1]))

        hue = _get_attr('hue')

        sat = _get_attr('sat')

        eulav = {
            'hue': hsb[0],
            'saturation': hsb[1]}

        crrnt = {
            'hue': int(hue),
            'saturation': int(sat)}

        if (crrnt == eulav
                and not force):
            return None


        quote = quote_plus(
            dumps(eulav))

        path = (
            f'devices/{unique}'
            f'/setColor/{quote}')

        if (change is False
                or homie.dryrun):
            return True

        return request(
            method='get',
            path=path,
            timeout=timeout)


    def _set_level() -> _RESPONSE:

        assert isinstance(
            value, int | float)

        path = (
            f'devices/{unique}'
            f'/setLevel/{value}')

        crrnt = _get_attr('level')

        if (crrnt == str(value)
                and not force):
            return None

        if (change is False
                or homie.dryrun):
            return True

        return request(
            method='get',
            path=path,
            timeout=timeout)


    if property == 'state':
        return _set_state()

    if property == 'color':
        return _set_color()

    if property == 'level':
        return _set_level()


    raise UnexpectedCondition



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from json import loads
from typing import TYPE_CHECKING

from encommon.types import DictStrAny
from encommon.utils import load_sample
from encommon.utils import prep_sample
from encommon.utils import read_text
from encommon.utils import rvrt_sample
from encommon.utils.sample import ENPYRWS

from . import SAMPLES
from ..origin import PhueOrigin
from ..surgeon import surgeon

if TYPE_CHECKING:
    from ...homie import Homie
    from ...utils import TestBodies



def test_surgeon(
    homie: 'Homie',
    replaces: DictStrAny,
    bodies: 'TestBodies',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    :param replaces: Mapping of what to replace in samples.
    :param bodies: Locations and groups for use in testing.
    """

    childs = homie.childs
    origins = childs.origins

    samples = SAMPLES / 'stream'


    planets = bodies.planets

    for planet in planets:

        origin = origins[
            f'{planet}_philips']

        assert isinstance(
            origin, PhueOrigin)

        assert origin.refresh()

        assert origin.fetch

        fetch = origin.fetch


        source = read_text(
            f'{samples}/dumped'
            f'/{planet}.json')

        source = rvrt_sample(
            sample=source,
            replace=replaces)

        sample = loads(source)


        for loaded in sample:

            event = loaded['event']

            assert surgeon(
                fetch, event)


        sample_path = (
            f'{SAMPLES}/surgeon'
            f'/{planet}.json')

        sample = load_sample(
            path=sample_path,
            update=ENPYRWS,
            content=fetch,
            replace=replaces)

        expect = prep_sample(
            content=fetch,
            replace=replaces)

        assert expect == sample



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from .aspired import HomieAspired
from .aspired import HomieAspiredItem
from .desired import HomieDesired
from .desired import HomieDesiredItem
from .persist import HomiePersist
from .persist import HomiePersistExpire
from .persist import HomiePersistValue



__all__ = [
    'HomieAspired',
    'HomieAspiredItem',
    'HomieDesired',
    'HomieDesiredItem',
    'HomiePersist',
    'HomiePersistValue',
    'HomiePersistExpire']



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from pathlib import Path
from time import sleep as block_sleep
from typing import TYPE_CHECKING

from encommon.types import DictStrAny
from encommon.types import inrepr
from encommon.types import instr
from encommon.types import lattrs
from encommon.utils import save_text

from respx import MockRouter

from ..desired import HomieDesiredItem
from ....conftest import config_factory
from ....conftest import homie_factory

if TYPE_CHECKING:
    from ...homie import Homie
    from ....utils import TestTimes



def test_HomieDesiredItem(
    homie: 'Homie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    """

    childs = homie.childs
    devices = childs.devices
    groups = childs.groups
    desires = childs.desires

    device = devices[
        'jupiter_light1']

    group1 = groups['ganymede']
    group2 = groups['halimede']

    desire = desires[
        'jupiter_active']

    model = HomieDesiredItem


    item = model(
        group1, desire)


    attrs = lattrs(item)

    assert attrs == [
        'target',
        'weight',
        'desire',
        'state',
        'color',
        'level',
        'scene']


    assert item.target

    assert item.desire

    assert item.state == 'poweron'

    assert not item.color

    assert item.level == 100

    assert item.scene


    item1 = model(
        group2, desire)

    item2 = model(
        device, desire)


    assert item < item1
    assert item < item2



def test_HomieDesired(
    homie: 'Homie',
    times: 'TestTimes',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    :param times: Times that are used with tests in project.
    """


    desired = homie.desired


    attrs = lattrs(desired)

    assert attrs == [
        '_HomieDesired__homie']


    assert inrepr(
        'desired.HomieDesired',
        desired)

    assert hash(desired) > 0

    assert instr(
        'desired.HomieDesired',
        desired)


    assert homie.refresh()


    time = times.middle

    aitems = (
        desired.items(time))

    assert len(aitems) == 0

    block_sleep(1.1)

    aitems = (
        desired.items(time))

    assert len(aitems) == 4



def test_HomieDesired_cover(
    tmp_path: Path,
    respx_mock: MockRouter,
    replaces: DictStrAny,
    times: 'TestTimes',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param tmp_path: pytest object for temporal filesystem.
    :param respx_mock: Object for mocking request operation.
    :param replaces: Mapping of what to replace in samples.
    :param times: Times that are used with tests in project.
    """

    samples = (
        tmp_path / 'homie')

    samples.mkdir()

    save_text(
        samples / 'test.yml',
        """
        desires:
         pytest:
          wheres:
          - builtins_period:
             start: '00:00'
        """)  # noqa: LIT003

    homie = homie_factory(
        config_factory(tmp_path),
        respx_mock=respx_mock)


    desired = homie.desired


    assert homie.refresh()


    time = times.middle

    aitems = (
        desired.items(time))

    assert len(aitems) == 0

    block_sleep(1.1)

    aitems = (
        desired.items(time))

    assert len(aitems) == 4



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from .action import HomieAction
from .action import HomieActionBase
from .action import HomieActionItem
from .action import HomieActionNode
from .stream import HomieStream
from .stream import HomieStreamItem
from .thread import HomieThread
from .thread import HomieThreadItem
from .update import HomieUpdate
from .update import HomieUpdateBase
from .update import HomieUpdateItem



__all__ = [
    'HomieThread',
    'HomieThreadItem',
    'HomieAction',
    'HomieActionBase',
    'HomieActionItem',
    'HomieActionNode',
    'HomieStream',
    'HomieStreamItem',
    'HomieUpdate',
    'HomieUpdateBase',
    'HomieUpdateItem']



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

from ..helpers import phue_changed
from ..helpers import phue_sensors
from ...origin import PhueOrigin
from ....utils.tests import STARTED

if TYPE_CHECKING:
    from ....homie import Homie
    from ....utils import TestBodies



def test_phue_changed(
    homie: 'Homie',
    bodies: 'TestBodies',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    :param bodies: Locations and groups for use in testing.
    """

    childs = homie.childs
    origins = childs.origins
    devices = childs.devices


    planets = bodies.planets

    for planet in planets:

        origin = origins[
            f'{planet}_philips']

        device = devices[
            f'{planet}_button']

        assert isinstance(
            origin, PhueOrigin)

        assert origin.refresh()


        timestamp = (
            STARTED
            .shift('-1d@h'))


        assert device.source

        changed = phue_changed(
            device.source)

        assert changed == {
            'button1': None,
            'button2': timestamp,
            'button3': timestamp,
            'button4': None}



def test_phue_sensors(
    homie: 'Homie',
    bodies: 'TestBodies',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    :param bodies: Locations and groups for use in testing.
    """

    childs = homie.childs
    origins = childs.origins
    devices = childs.devices


    planets = bodies.planets

    for planet in planets:

        origin = origins[
            f'{planet}_philips']

        device = devices[
            f'{planet}_motion']

        assert isinstance(
            origin, PhueOrigin)

        assert origin.refresh()


        assert device.source

        sensors = phue_sensors(
            device.source)

        prefix = (
            'aa012345'
            if planet == 'jupiter'
            else 'bb012345')

        motion = (
            f'{prefix}-abcd-1234'
            '-ab12-abcdef000025')

        assert sensors == {
            'motion': motion}



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from ..models import BltnModels



def test_BltnModels_cover() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    models = BltnModels

    drivers = models.drivers()

    assert drivers.store()
    assert drivers.period()
    assert drivers.regexp()



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from pathlib import Path

from encommon.types import DictStrAny
from encommon.utils import load_sample
from encommon.utils import prep_sample
from encommon.utils.sample import ENPYRWS

from . import SAMPLES
from ..config import HomieConfig



def test_HomieConfig(
    tmp_path: Path,
    config: HomieConfig,
    replaces: DictStrAny,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param tmp_path: pytest object for temporal filesystem.
    :param replaces: Mapping of what to replace in samples.
    :param config: Primary class instance for configuration.
    """

    sample_path = (
        SAMPLES / 'config.json')

    sample = load_sample(
        path=sample_path,
        update=ENPYRWS,
        content=config.config,
        replace=replaces)

    expect = prep_sample(
        content=config.config,
        replace=replaces)

    assert expect == sample



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from .origin import PhueOriginParams
from .plugins import DriverPhueButtonParams
from .plugins import DriverPhueChangeParams
from .plugins import DriverPhueContactParams
from .plugins import DriverPhueMotionParams
from .plugins import DriverPhueSceneParams



__all__ = [
    'PhueOriginParams',
    'DriverPhueButtonParams',
    'DriverPhueContactParams',
    'DriverPhueMotionParams',
    'DriverPhueChangeParams',
    'DriverPhueSceneParams']



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from re import compile
from typing import Annotated
from typing import Any
from typing import Literal
from typing import Optional

from encommon.types import BaseModel

from pydantic import Field



_WEEKDAY = Literal[
    'Monday',
    'Tuesday',
    'Wednesday',
    'Thursday',
    'Friday',
    'Saturday',
    'Sunday']

_TIMERGXP = compile(
    r'^([01]\d|2[0-3])'
    r':([0-5]\d)$')



class DriverBltnPeriodParams(BaseModel, extra='forbid'):
    """
    Process and validate the Homie configuration parameters.
    """

    start: Annotated[
        Optional[str],
        Field(None,
              description='Start time in 24 hour format',
              pattern=_TIMERGXP)]

    stop: Annotated[
        Optional[str],
        Field(None,
              description='Stop time in 24 hour format',
              pattern=_TIMERGXP)]

    days: Annotated[
        Optional[list[_WEEKDAY]],
        Field(None,
              description='Days of the week in scope',
              min_length=1)]

    tzname: Annotated[
        str,
        Field('UTC',
              description='Timezone parsed with Times',
              min_length=1)]


    def __init__(
        # NOCVR
        self,
        **data: Any,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        days = data.get('days')

        if isinstance(days, str):
            data['days'] = [days]

        super().__init__(**data)



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING
from typing import Union

from encommon.times import Times

if TYPE_CHECKING:
    from .aspire import HomieAspire
    from .desire import HomieDesire
    from ..threads import HomieStreamItem



_OCCURD_WHENS = list[bool]

_WHERE_WHENS = dict[str, bool]
_WHERE_WHENL = dict[str, list[bool]]



def whered(
    desire: Union['HomieDesire', 'HomieAspire'],
    time: Times,
) -> tuple[bool, ...]:
    """
    Return the boolean indicating the conditional outcomes.

    :param desire: Child class instance for Homie Automate.
    :param time: Time that will be used in the conditionals.
    :returns: Boolean indicating the conditional outcomes.
    """

    plugins = desire.wheres

    wheres: _WHERE_WHENS = {}


    lists: _WHERE_WHENL = {}

    for plugin in plugins:

        params = plugin.params
        negate = params.negate
        family = params.family

        if family not in lists:
            lists[family] = []

        _lists = lists[family]

        when = plugin.when(time)

        if negate is True:
            when = not when

        _lists.append(when)


    items = lists.items()

    for name, values in items:

        value = any(values)

        if name == 'default':
            value = all(values)

        wheres[name] = value


    _wheres = wheres.values()

    return tuple(_wheres)



def occurd(
    aspire: 'HomieAspire',
    sitem: 'HomieStreamItem',
) -> tuple[bool, ...]:
    """
    Return the boolean indicating the conditional outcomes.

    :param aspire: Child class instance for Homie Automate.
    :param sitem: Item containing information for operation.
    :returns: Boolean indicating the conditional outcomes.
    """

    plugins = aspire.occurs

    occurs: _OCCURD_WHENS = []


    for plugin in plugins:

        when = plugin.when(sitem)

        occurs.append(when)


    return tuple(occurs)



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Annotated

from encommon.types import BaseModel

from pydantic import Field



class HomieServiceTimeoutParams(BaseModel, extra='forbid'):
    """
    Process and validate the Homie configuration parameters.
    """

    action: Annotated[
        int,
        Field(5,
              description='Override the actions timeout',
              ge=1, le=30)]

    update: Annotated[
        int,
        Field(3,
              description='Override the updates timeout',
              ge=1, le=30)]

    stream: Annotated[
        int,
        Field(60,
              description='Override the stream timeout',
              ge=1, le=900)]



class HomieServiceRespiteParams(BaseModel, extra='forbid'):
    """
    Process and validate the Homie configuration parameters.
    """

    desire: Annotated[
        int,
        Field(60,
              description='How often desires performed',
              ge=1, le=900)]

    update: Annotated[
        int,
        Field(60,
              description='How often updates performed',
              ge=1, le=900)]

    health: Annotated[
        int,
        Field(3,
              description='How often health is checked',
              ge=1, le=15)]



class HomieServiceParams(BaseModel, extra='forbid'):
    """
    Process and validate the Homie configuration parameters.
    """

    respite: Annotated[
        HomieServiceRespiteParams,
        Field(default_factory=HomieServiceRespiteParams,
              description='When operates are performed')]

    timeout: Annotated[
        HomieServiceTimeoutParams,
        Field(default_factory=HomieServiceTimeoutParams,
              description='Override source or defaults')]



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from pathlib import Path



SAMPLES = (
    Path(__file__).parent
    / 'samples')



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from copy import deepcopy
from pathlib import Path

from encommon.types import DictStrAny
from encommon.utils import save_text

from pytest import fixture

from respx import MockRouter

from . import EXAMPLES
from . import PROJECT
from .homie import Homie
from .homie import HomieConfig
from .homie import HomieService
from .hubitat.test import mock_hubi
from .philips.test import mock_phue
from .ubiquiti.test import mock_ubiq
from .utils import TestBodies
from .utils import TestTimes



def config_factory(
    tmp_path: Path,
) -> HomieConfig:
    """
    Construct the instance for use in the downstream tests.

    :param tmp_path: pytest object for temporal filesystem.
    :returns: Newly constructed instance of related class.
    """

    content = (
        f"""

        enconfig:
          paths:
            - {EXAMPLES}
            - {tmp_path}/homie

        enlogger:
          stdo_level: info

        database: >-
            sqlite:///{tmp_path}/db

        """)  # noqa: LIT003

    config_path = (
        tmp_path / 'config.yml')

    save_text(
        config_path, content)

    sargs = {
        'config': config_path,
        'dryrun': False,
        'potent': True,
        'console': True,
        'debug': True,
        'idesire': 1,
        'iupdate': 1,
        'ihealth': 1,
        'atimeout': 1,
        'utimeout': 1,
        'stimeout': 1,
        'paction': True,
        'pupdate': True,
        'pstream': True,
        'paspire': True,
        'pdesire': True}

    return HomieConfig(sargs)



@fixture
def config(
    tmp_path: Path,
) -> HomieConfig:
    """
    Construct the instance for use in the downstream tests.

    :param tmp_path: pytest object for temporal filesystem.
    :returns: Newly constructed instance of related class.
    """

    return config_factory(tmp_path)



@fixture
def times() -> TestTimes:
    """
    Return the simple mapping of what to replace in sample.

    :returns: Simple mapping of what to replace in sample.
    """

    return TestTimes()



@fixture
def bodies() -> TestBodies:
    """
    Return the simple mapping of what to replace in sample.

    :returns: Simple mapping of what to replace in sample.
    """

    return TestBodies()



def replaces_factory() -> DictStrAny:
    """
    Return the complete mapping of what to replace in sample.

    :returns: Complete mapping of what to replace in sample.
    """

    times = TestTimes()

    replaces = {

        'PROJECT': PROJECT,

        '__TIMESTAMP_START__': (
            times.start.simple),

        '__TIMESTAMP_MIDDLE__': (
            times.middle.simple),

        '__TIMESTAMP_FINAL__': (
            times.final.simple),

        '__TIMESTAMP_CRRNT__': (
            times.current.simple),

        '__UNIXEPOCH_START__': (
            times.start.spoch),

        '__UNIXEPOCH_MIDDLE__': (
            times.middle.spoch),

        '__UNIXEPOCH_FINAL__': (
            times.final.spoch),

        '__UNIXEPOCH_CRRNT__': (
            times.current.spoch)}

    return deepcopy(replaces)



@fixture
def replaces(
    tmp_path: Path,
) -> DictStrAny:
    """
    Return the complete mapping of what replaced in sample.

    :param tmp_path: pytest object for temporal filesystem.
    :returns: Complete mapping of what replaced in sample.
    """

    replaces = replaces_factory()

    return replaces | {
        'TMPPATH': tmp_path}



def homie_factory(
    config: HomieConfig,
    respx_mock: MockRouter,
) -> Homie:
    """
    Construct the instance for use in the downstream tests.

    :param config: Primary class instance for configuration.
    :param respx_mock: Object for mocking request operation.
    :returns: Newly constructed instance of related class.
    """

    replaces = replaces_factory()

    mock_hubi(respx_mock, replaces)
    mock_phue(respx_mock, replaces)
    mock_ubiq(respx_mock, replaces)

    return Homie(config)



@fixture
def homie(
    config: HomieConfig,
    respx_mock: MockRouter,
) -> Homie:
    """
    Construct the instance for use in the downstream tests.

    :param config: Primary class instance for configuration.
    :param respx_mock: Object for mocking request operation.
    :returns: Newly constructed instance of related class.
    """

    return homie_factory(
        config, respx_mock)



def service_factory(
    homie: Homie,
    respx_mock: MockRouter,
) -> HomieService:
    """
    Construct the instance for use in the downstream tests.

    :param homie: Primary class instance for Homie Automate.
    :param respx_mock: Object for mocking request operation.
    :returns: Newly constructed instance of related class.
    """

    return HomieService(homie)



@fixture
def service(
    homie: Homie,
    respx_mock: MockRouter,
) -> HomieService:
    """
    Construct the instance for use in the downstream tests.

    :param homie: Primary class instance for Homie Automate.
    :param respx_mock: Object for mocking request operation.
    :returns: Newly constructed instance of related class.
    """

    return service_factory(
        homie, respx_mock)



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Literal



HomieChildKinds = Literal[
    'origin',
    'device',
    'group',
    'scene',
    'aspire',
    'desire']



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

from encommon.types import inrepr
from encommon.types import instr
from encommon.types import lattrs

from ..queue import HomieQueue
from ..queue import HomieQueueItem

if TYPE_CHECKING:
    from ..homie import Homie



def test_HomieQueueItem() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    item = HomieQueueItem()

    assert item.time.since < 1



def test_HomieQueue(
    homie: 'Homie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    """

    _queue = HomieQueue[int]

    queue: _queue = (
        HomieQueue(homie))


    attrs = lattrs(queue)

    assert attrs == [
        '_HomieQueue__homie',
        '_HomieQueue__queue']


    assert inrepr(
        'queue.HomieQueue',
        queue)

    assert hash(queue) > 0

    assert instr(
        'queue.HomieQueue',
        queue)



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from pathlib import Path
from typing import TYPE_CHECKING

from encommon.types import inrepr
from encommon.types import instr
from encommon.types import lattrs
from encommon.utils import save_text

from respx import MockRouter

from ....conftest import config_factory
from ....conftest import homie_factory

if TYPE_CHECKING:
    from ...homie import Homie
    from ....utils import TestBodies



def test_HomieAspire(
    homie: 'Homie',
    bodies: 'TestBodies',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    :param bodies: Locations and groups for use in testing.
    """

    childs = homie.childs
    aspires = childs.aspires


    planets = bodies.planets

    for planet in planets:

        name = f'{planet}_active'

        aspire = aspires[name]


        attrs = lattrs(aspire)

        assert attrs == [
            '_HomieChild__homie',
            '_HomieChild__name',
            '_HomieChild__params',
            '_HomieAspire__occurs',
            '_HomieAspire__wheres',
            '_HomieAspire__timer']


        assert inrepr(
            'aspire.HomieAspire',
            aspire)

        assert hash(aspire) > 0

        assert instr(
            'aspire.HomieAspire',
            aspire)


        aspire.validate()

        assert aspire.homie

        assert aspire.name == name

        assert aspire.family == 'builtins'

        assert aspire.kind == 'aspire'

        assert aspire.params

        assert len(aspire.devices) == 1

        assert len(aspire.groups) == 1

        assert len(aspire.occurs) == 6

        assert len(aspire.wheres) == 1

        assert not aspire.paused

        assert aspire.dumped



def test_HomieAspire_cover(
    tmp_path: Path,
    respx_mock: MockRouter,
    bodies: 'TestBodies',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param tmp_path: pytest object for temporal filesystem.
    :param respx_mock: Object for mocking request operation.
    :param bodies: Locations and groups for use in testing.
    """

    samples = (
        tmp_path / 'homie')

    samples.mkdir()

    save_text(
        samples / 'test.yml',
        """
        aspires:
         pytest:
          store:
          - unique: aspire
            value: true
            expire: '1y'
        """)  # noqa: LIT003

    homie = homie_factory(
        config_factory(tmp_path),
        respx_mock=respx_mock)


    childs = homie.childs
    origins = childs.origins

    aspire = (
        childs.aspires
        ['pytest'])


    assert not aspire.devices

    assert not aspire.groups


    planets = bodies.planets

    event = {'foo': 'bar'}

    for planet in planets:

        origin = origins[
            f'{planet}_philips']

        sitem = (
            origin
            .get_stream(event))

        assert not (
            aspire.when(sitem))



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Literal
from typing import TYPE_CHECKING
from typing import Union

if TYPE_CHECKING:
    from ..homie.childs import HomieChild



_PHASE = Literal[
    'initial',
    'runtime']



class InvalidChild(Exception):
    """
    Exception for when the child could not be instantiated.

    :param child: Name or child that is determined invalid.
    :param phase: From which phase child was found invalid.
    """

    child: str
    phase: _PHASE


    def __init__(
        self,
        child: Union[str, 'HomieChild'],
        phase: _PHASE,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        if hasattr(child, 'name'):
            child = child.name

        message = (
            f'Child ({child}) '
            'invalid within '
            f'phase ({phase})')

        self.child = child

        super().__init__(message)



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from json import loads
from pathlib import Path
from typing import TYPE_CHECKING

from encommon.types import DictStrAny
from encommon.types import inrepr
from encommon.types import instr
from encommon.types import lattrs
from encommon.utils import read_text
from encommon.utils import rvrt_sample
from encommon.utils import save_text

from respx import MockRouter

from ..scene import DriverPhueScene
from ...test import SAMPLES
from ....conftest import config_factory
from ....conftest import homie_factory

if TYPE_CHECKING:
    from ....homie import Homie
    from ....utils import TestBodies



def test_DriverPhueScene(
    homie: 'Homie',
    replaces: DictStrAny,
    bodies: 'TestBodies',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    :param replaces: Mapping of what to replace in samples.
    :param bodies: Locations and groups for use in testing.
    """

    childs = homie.childs
    origins = childs.origins
    aspires = childs.aspires


    assert homie.refresh()


    planets = bodies.planets

    family = 'philips'

    for planet in planets:

        origin = origins[
            f'{planet}_{family}']

        aspire = aspires[
            f'{planet}_active']

        driver = (
            aspire.occurs[4]
            .drivers[0])

        assert isinstance(
            driver, DriverPhueScene)


        attrs = lattrs(driver)

        assert attrs == [
            '_HomieDriver__plugin',
            '_HomieDriver__params']


        assert inrepr(
            'scene.DriverPhueScene',
            driver)

        assert hash(driver) > 0

        assert instr(
            'scene.DriverPhueScene',
            driver)


        driver.validate()

        assert driver.plugin

        assert driver.family == family

        assert driver.kinds == ['occur']

        assert driver.params


        samples = SAMPLES / 'stream'

        source = read_text(
            f'{samples}/dumped'
            f'/{planet}.json')

        source = rvrt_sample(
            sample=source,
            replace=replaces)

        sample = loads(source)


        sitem = (
            origin.get_stream(
                sample[7]['event']))

        assert driver.occur(sitem)



def test_DriverPhueScene_cover(
    tmp_path: Path,
    respx_mock: MockRouter,
    replaces: DictStrAny,
    bodies: 'TestBodies',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param tmp_path: pytest object for temporal filesystem.
    :param respx_mock: Object for mocking request operation.
    :param replaces: Mapping of what to replace in samples.
    :param bodies: Locations and groups for use in testing.
    """

    samples = (
        tmp_path / 'homie')

    samples.mkdir()

    save_text(
        samples / 'test.yml',
        """
        groups:
         jupiter_nexist:
          origin: jupiter_philips
          label: Not exists
         jupiter_norigin:
          devices: jupiter_special
         neptune_nexist:
          origin: neptune_philips
          label: Not exists
         neptune_norigin:
          devices: neptune_special
        aspires:
         pytest:
          groups:
          - ganymede
          - halimede
          occurs:
          - philips_scene:
             scene: active
             group: jupiter_nexist
             states: active
          - philips_scene:
             scene: active
             group: jupiter_norigin
             states: active
          - philips_scene:
             scene: active
             group: neptune_nexist
             states: active
          - philips_scene:
             scene: active
             group: neptune_norigin
             states: active
        """)  # noqa: LIT003

    homie = homie_factory(
        config_factory(tmp_path),
        respx_mock=respx_mock)


    childs = homie.childs
    origins = childs.origins

    origin = origins[
        'jupiter_philips']

    aspire = (
        childs.aspires
        ['pytest'])

    samples = SAMPLES / 'stream'


    assert homie.refresh()


    planets = bodies.planets

    loaded: list[DictStrAny] = []

    for planet in planets:

        source = read_text(
            f'{samples}/dumped'
            f'/{planet}.json')

        source = rvrt_sample(
            sample=source,
            replace=replaces)

        sample = loads(source)

        loaded.extend(sample)


    for event in loaded:

        sitem = (
            origin.get_stream(
                event['event']))

        occurs = aspire.occurs

        for occur in occurs:

            (occur.drivers[0]
             .occur(sitem))



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

from encommon.types import inrepr
from encommon.types import instr
from encommon.types import lattrs

from ..child import InvalidChild

if TYPE_CHECKING:
    from ...homie import Homie



def test_InvalidChild() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    raises = InvalidChild(
        child='invalid',
        phase='initial')


    attrs = lattrs(raises)

    assert attrs == ['child']


    assert inrepr(
        'InvalidChild',
        raises)

    assert hash(raises) > 0

    assert instr(
        'Child (invalid)',
        raises)


    assert str(raises) == (
        'Child (invalid) '
        'invalid within '
        'phase (initial)')



def test_InvalidChild_cover(
    homie: 'Homie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    """

    childs = homie.childs
    devices = childs.devices

    device = devices[
        'jupiter_motion']


    raises = InvalidChild(
        child=device,
        phase='runtime')


    assert str(raises) == (
        'Child (jupiter_motion) '
        'invalid within '
        'phase (runtime)')



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from dataclasses import asdict
from typing import TYPE_CHECKING

from encommon.types import DictStrAny
from encommon.utils import load_sample
from encommon.utils import prep_sample
from encommon.utils.sample import ENPYRWS

from . import SAMPLES
from ..stream import PhueStreamItem

if TYPE_CHECKING:
    from ...homie import HomieService
    from ...utils import TestBodies



def test_PhueStreamItem(
    service: 'HomieService',
    bodies: 'TestBodies',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param service: Ancilary Homie Automate class instance.
    :param bodies: Locations and groups for use in testing.
    """

    homie = service.homie
    childs = homie.childs
    origins = childs.origins

    model = PhueStreamItem


    planets = bodies.planets

    event = {'foo': 'bar'}

    for planet in planets:

        name = f'{planet}_philips'

        origin = origins[name]


        item = model(
            origin, event)


        assert item.origin == name

        assert item.event == event



def test_PhueStream_samples(
    service: 'HomieService',
    replaces: DictStrAny,
    bodies: 'TestBodies',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param service: Ancilary Homie Automate class instance.
    :param replaces: Mapping of what to replace in samples.
    :param bodies: Locations and groups for use in testing.
    """

    member = service.streams
    threads = member.threads
    squeue = member.squeue
    cancel = member.cancel

    samples = SAMPLES / 'stream'


    planets = bodies.planets

    for planet in planets:

        cancel.clear()

        thread = threads[
            f'{planet}_philips']

        thread.operate()

        assert squeue.qsize == 8

        cancel.set()

        thread.operate()

        assert squeue.qsize == 8


        output: list[DictStrAny] = []

        while not squeue.empty:

            sitem = squeue.get()

            append = asdict(sitem)

            del append['time']

            output.append(append)


        sample_path = (
            f'{samples}/dumped'
            f'/{planet}.json')

        sample = load_sample(
            path=sample_path,
            update=ENPYRWS,
            content=output,
            replace=replaces)

        expect = prep_sample(
            content=output,
            replace=replaces)

        assert expect == sample



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

from encommon.types import inrepr
from encommon.types import instr
from encommon.types import lattrs

if TYPE_CHECKING:
    from ...homie import Homie



def test_HomieChilds(
    homie: 'Homie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    """

    childs = homie.childs


    attrs = lattrs(childs)

    assert attrs == [
        '_HomieChilds__homie',
        '_HomieChilds__origins',
        '_HomieChilds__devices',
        '_HomieChilds__groups',
        '_HomieChilds__scenes',
        '_HomieChilds__desires',
        '_HomieChilds__aspires']


    assert inrepr(
        'homie.HomieChilds',
        childs)

    assert hash(childs) > 0

    assert instr(
        'homie.HomieChilds',
        childs)


    childs.validate()

    assert childs.origins

    assert childs.devices

    assert childs.groups

    assert childs.scenes

    assert childs.desires

    assert childs.aspires



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

from ..helpers import ubiq_latest
from ...origin import UbiqOrigin

if TYPE_CHECKING:
    from ....homie import Homie
    from ....utils import TestBodies



def test_ubiq_latest(
    homie: 'Homie',
    bodies: 'TestBodies',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    :param bodies: Locations and groups for use in testing.
    """

    childs = homie.childs
    origins = childs.origins


    planets = bodies.planets

    for planet in planets:

        origin = origins[
            f'{planet}_ubiquiti']

        assert isinstance(
            origin, UbiqOrigin)

        assert origin.refresh()


        assert origin.merge

        values = (
            origin.merge
            .values())


        for value in values:

            latest = ubiq_latest(value)

            assert latest > '-2d'



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING
from typing import Type

if TYPE_CHECKING:
    from .action import HubiActionItem
    from .params import HubiOriginParams
    from .update import HubiUpdateItem



class HubiModels:
    """
    Return the class object that was imported within method.
    """


    @classmethod
    def origin(
        cls,
    ) -> Type['HubiOriginParams']:
        """
        Return the class object that was imported within method.

        :returns: Class object that was imported within method.
        """

        from .params import (
            HubiOriginParams)

        return HubiOriginParams


    @classmethod
    def action(
        cls,
    ) -> Type['HubiActionItem']:
        """
        Return the class object that was imported within method.

        :returns: Class object that was imported within method.
        """

        from .action import (
            HubiActionItem)

        return HubiActionItem


    @classmethod
    def update(
        cls,
    ) -> Type['HubiUpdateItem']:
        """
        Return the class object that was imported within method.

        :returns: Class object that was imported within method.
        """

        from .update import (
            HubiUpdateItem)

        return HubiUpdateItem



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from .origin import UbiqOriginParams
from .plugins import DriverUbiqClientParams



__all__ = [
    'UbiqOriginParams',
    'DriverUbiqClientParams']



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from .period import DriverBltnPeriodParams
from .regexp import DriverBltnRegexpParams
from .store import DriverBltnStoreParams



__all__ = [
    'DriverBltnStoreParams',
    'DriverBltnPeriodParams',
    'DriverBltnRegexpParams']
